'*******************************************************************************
'*                                                                             *
'* ZOG      A ZPU Virtual machine running on the Parallax Propeller            *
'*          micro-controller.                                                  *
'*                                                                             *
'* What?    The ZPU is a 32 bit processor core architecture specified by ZyLin *
'*          consultants for which there a number of HDL implementations        *
'*          for FPGA's.                                                        *
'*                                                                             *
'* Why?     ZyLin have targeted the GNU C Compiler to the ZPU instruction set. *
'*          ZOG makes it possible to use GCC for Propeller development.        *
'*          All ZPU instructions are a single byte, which should work out      *
'*          nicely when the code is placed in external RAM with an 8 bit bus.  *
'*                                                                             *
'*          Based on an original idea by Toby Seckshund.                       *
'*          Maths routines courtesy of Cluso and Parallax.                     *
'*                                                                             *
'*          Encouragement courtesy of Bill Henning and all on the Parallax     *
'*          Propeller discussion forum.                                        *
'*                                                                             *
'* Author:  Michael Rychlik                                                    *
'*                                                                             *
'*******************************************************************************
'
' v0.1      09-02-2010 First draft, totally untested.
'
' v0.2      18-02-2010 Tested and fixed up initial bugs.
'                      Runs first test program compiled with GCC.
'                      Has no I/O yet.
'
' v0.3      19-02-2010 Fixed mysterious "inversion of bit four of offset" problem
'                      with LOADSP and STORESP.
'
' v0.4      20-02-2010 Added primitive UART output such that the C run time
'                      routine outbyte() works. Zog now says hello!
'
' v0.5      28-02-2010 The CONFIG instruction is now always emulated,
'                      see emulate code in libgloss crt0.S for why.
'                      Added handling of the SYSCALL instruction via the io_command trap mechainism.
'                      Added fsrwFemto object (inititialization calls commented out for now).
'
' v0.6      xx-03-2010 Changed UART object to FullDuplexSerialPlus.
'
' v0.8      18-04-2010 Rewrote all ops with optimizations as suggested by Bill Henning.
'
' v0.9      25-04-2010 Optimized some most used ops in the FIBO test (wink, wink:)
'                      Fixed PUSHSP.
'
' v0.11     09-05-2010 Changes endianess of bytecodes in zpu_memory
'                      Adjusted read/write byte/word and instruction fetch inline with endianness
'                      Fixed ADDSP.
'
' v0.12     11-05-2010 The ZPU_EMULATE op now works. If it's ever needed.
'                      Fixed MOD and DIV instructions.
'                      Inlined read/write_byte
'                      read/write_word are EMULLATED for now as their endiannes is not right yet.
'                      ZPU memory is 20K bytes (Until we put SD and VM COG back in).

' v0.13     14-05-2010 Fixed endiannes of strings output by SYS_write syscall.
'                      Return a zero result from SYS_fstat syscall.
'
' v0.14     08-06-2010 Take Bill Henning's VMCog into use for external RAM access.
'                      Added #ifdef USE_VIRTUAL_MEMORY
'
' v0.15     15-06-2010 Fixed zpu_addsp to work correctly with virtual mem read_long.
'                      Added reading of a ZPU image from SD card file.
'                      Runs dhrystone by default (dstone.bin)
'
' v0.16     ??-06-2010 Fixed zpu_emulate for virtual memory usage.
'                      Take VMCog v0.975 into use.
'
' v0.17     20-06-2010 Ensure still builds for HUB memory use.
'
' v0.18     23-06-2010 Test for IM instruction in execute loop and halve the dispatch table size.
'                      Moved clearing of decode_mask into execute loop, saves many LONGs
'                      Change testing of decode_mask to use tjz
'                      Removed FIBO optimization code. We don't like to cheat.
'                      Changed dispatch table entries to BYTES, halving it's size again
'                      and saving one instruction in the exececute loop.
'                      Simplified checking for I/O in read/write_long
'                      These changes bring a 13% boost to fibo performance! (Running from HUB).
'
' v0.19     23-06-2010 read/write long now have access to HUB memory at virtual address $10000
'
' v0.20     01-07-2010 Now Zog C can start a cog via SYS_cognew syscall.
'
' v0.21     03-07-2010 Now runs new "pasm_objects_test" that tests C++ versions
'                      of FDS and VMCOG.
'
' v0.22     04-07-2010 Now handle SYS_cognew in PASM rather than Spin.
'
' v1.0      08-07-2010 Zog is now split out into it's own Spin file with minimal Spin
'                      startup code.
'                      Can now totally replace Spin and take over the entire Prop.
'                      N.B. Virtual Memory opertion is broken at this point.
'
' v1.1      17-08-2010 Fixed virtual memory operation.
'
' v1.2      17-08-2010 No changes here but all make files for ZPU executables now perform
'                      and endianness swap on the binaries as byte swapping is not done here anymore.
'
' v1.3      18-08-2010 No change but bump version inline with ZOG.
'
' v1.4      26-08-2010 Added display of "NOT IMPLEMENTED" ops in print_regs.
'
' v1.5      28-08-2010 Added support for Jazzed's SDRAM external RAM solution.
'                         ...define USE_JCACHED_MEMORY to use that.
'                      Adopted Jazzed's use of userdefs.spin for platform configuration.
'
' v1.6      31-08-2010 No change but bump version inline with ZOG.
'
' The "*.bin" file included in the ZPU memory area is created with
' GCC for the ZPU.
'
' Please see the Makefile included with Zog for how to build C programs
' with GCC for the ZPU architecture that Zog can run.
'
'#define SINGLE_STEP
'#define CYCLE_COUNT

'Select use of HUB memory or external memory here
#define USE_JCACHED_MEMORY
'#define USE_VIRTUAL_MEMORY
'#define USE_HUB_MEMORY

#ifdef USE_JCACHED_MEMORY
#define USE_SDCARD
#endif

#ifdef USE_VIRTUAL_MEMORY
#define USE_SDCARD
#endif

#ifdef USE_HUB_MEMORY
#endif

OBJ
'Set your platform configuration in userdef.spin
'Ideally, your userdefs.spin can live in the library path,
'and you can simply remove a new package's userdefs.spin
'to enable your own definitions.
    def : "userdefs.spin"

CON
    _clkmode        = def#_clkmode
    _xinfreq        = def#_xinfreq
    ' user baud rate
    baud            = def#conBaud
    ' user microSD pins
    spiDO           = def#spiDO
    spiClk          = def#spiCLK
    spiDI           = def#spiDI
    spiCLR          = def#spiCLR
    spiINC          = def#spiINC

DAT                     org 0
zpu_memory              byte ' Force zpu_memory to be BYTE type.
#ifdef USE_HUB_MEMORY
'zpu_image               file "test.bin"
'zpu_image               file "filetest.bin"
zpu_image               file "fibo.bin"
'zpu_image               file "mall.bin"
padding                 byte 0[(zpu_memory_size) - (@padding - @zpu_memory)]
zpu_memory_end
                        fit (zpu_memory_size / 4)
#endif
#ifdef USE_SDCARD
DAT
'filename byte "vm.zbn",0
'filename byte "xbasic.zbn",0
'filename byte "fibo.zbn",0
'filename byte "fibo.bin",0
filename byte "filetest.zbn",0
'filename byte "filetest.bin",0
'filename byte "mall.bin",0
'filename byte "hello.bin",0
#endif

CON
#ifdef USE_JCACHED_MEMORY
zpu_memory_size = (64 * 1024)                            'The size of ZPU memory area
'zpu_memory_size = (1 << 25)                              'The size of ZPU memory area
hub_memory_size = 32 * 1024
rendezvous_area = hub_memory_size - 128
scratch_buffer = rendezvous_area - 2048
vm_cache = scratch_buffer - 8192
vm_mbox = rendezvous_area
vm_params = vm_mbox + 8
serial_data = vm_params + 16
#endif
#ifdef USE_VIRTUAL_MEMORY
zpu_memory_size = (64 * 1024)                            'The size of ZPU memory area
#endif
#ifdef USE_HUB_MEMORY
zpu_memory_size = (26 * 1024)                            'The size of ZPU memory area
#endif

' These are the SPIN byte codes for mul and div
SPIN_MUL_OP     = $F4  '(multiply, return lower 32 bits)
SPIN_DIV_OP     = $F6  '(divide, return quotient 32 bits)
SPIN_REM_OP     = $F7  '(divide, return remainder 32 bits)

'I/O control block commands
io_cmd_out      = $01
io_cmd_in       = $02
io_cmd_break    = $03
io_cmd_syscall  = $04

'System  call ID numbers. Taken from libgloss syscall.h
'These are required by the ANSI C part of newlib (excluding system() of course).
SYS_exit        =  1
SYS_open        =  2
SYS_close       =  3
SYS_read        =  4
SYS_write       =  5
SYS_lseek       =  6
SYS_unlink      =  7
SYS_getpid      =  8
SYS_kill        =  9
SYS_fstat       = 10
'SYS_sbrk       = 11 - not currently a system call, but reserved.
'ARGV support.
SYS_argvlen     = 12
SYS_argv        = 13
'These are extras added for one reason or another.
SYS_chdir       = 14
SYS_stat        = 15
SYS_chmod       = 16
SYS_utime       = 17
SYS_time        = 18
'Propeller specific actions
SYS_cognew      = 8000
SYS_coginit     = 8001
SYS_cogstop     = 8002

'Standard file decriptors
STDIN_FILENO    = 0
STDOUT_FILENO   = 1
STDERR_FILENO   = 2

'Whence values for lseek
SEEK_SET        = 0
SEEK_CUR        = 1
SEEK_END        = 2

SECTORIO_FILENO = 3 ' hack to allow raw sector reads/writes

UART_TX_PORT  = $80000024 'ZPU virtual UART I/O ports
UART_RX_PORT  = $80000028

OBJ
  zog  : "zog"
  ser  : "FullDuplexSerialPlus"
#ifdef USE_SDCARD
  sd   : "fsrwFemto_rr001"                          'SD Software used in FemtoBASIC
#endif
#ifdef USE_JCACHED_MEMORY
  cm   : "spisram_cache_c3"                               'sdram cache interface
#endif
#ifdef USE_VIRTUAL_MEMORY
  vm   : "vmcog_c3"                        'The MMU/virtual memory manager cog
#endif

VAR
  '8 consecutive longs for Zog's par block.
  long par_zpu_memory_addr
  long par_zpu_memory_sz
  long par_initial_pc
  long par_initial_sp
  long par_dispatch_tab_addr
  long par_pasm_addr
  long par_zog_mbox_addr
  long par_vm_mbox_addr

  '8 consecutive longs for Zog's I/O mailbox
  'Need not follow on from par block above
  long zog_mbox_command
  long zog_mbox_port
  long zog_mbox_data
  long zog_mbox_pc
  long zog_mbox_sp
  long zog_mbox_tos
  long zog_mbox_dm
  long zog_mbox_debug

#ifdef USE_JCACHED_MEMORY
  'Mail box for Cache
  'Need not follow on from zog_mbox above
  'long vm_mbox[2] -- moved to rendezvous area
#endif

#ifdef USE_VIRTUAL_MEMORY
  'Mail box for Virtual Memory
  'Need not follow on from zog_mbox above
  long vm_mbox[4]
#endif

#ifdef USE_SDCARD
  long file_offset
#endif

  'Misc vars
  long cog
  long ioControl[2]               'SD parameters
  long framep
  word err                        'Error code return
  long step_count
  byte diskbuff[512]

PUB start : okay | n
  ser.start(def#conRxPin, def#conTxPin, def#conMode, def#conBaud) 'Start the debug Terminal
  ser.str(string("ZOG v1.6"))

#ifdef USE_JCACHED_MEMORY
  cm.start(vm_mbox, vm_cache)                       'Start up cache
  ser.str(string(" (CACHE)"))
#endif

#ifdef USE_VIRTUAL_MEMORY
  vm.start(@vm_mbox, $7C00, 10)                         'Start up VMCOG
  ser.str(string(" (VM)"))
#endif

#ifdef USE_HUB_MEMORY
  ser.str(string(" (HUB)", 13))
  ser.str(string("zpu memory at "))
  ser.hex(@zpu_memory, 8)
  par_initial_pc := 0
#endif
  crlf
  waitcnt(cnt+80_000_000)                               'Give serial terminal window time

#ifdef USE_SDCARD
  ser.str (string("Starting SD driver...",))
  err := sd.startSD(@ioControl)                         'Start the SD routines
  ser.hex(err, 8)
  crlf

  repeat
    ser.str (string("Mounting SD...",))
    err := \sd.mount_explicit(spiDO,spiClk,spiDI,spiINC)    'Mount the SD
    ser.hex(err, 8)
    crlf
  until err == 0

  file_offset := 0

  'load_bytecode
  write_bytecode_to_flash
  start_bytecode_in_flash
  'check_bytecode
  'dump_virtual
  'prime_vm
#endif

  'Set up Zog's parameter block
  par_zpu_memory_addr := @zpu_memory
  par_zpu_memory_sz := zpu_memory_size
  'par_initial_pc := 0
  par_initial_sp := zpu_memory_size - 8
  par_dispatch_tab_addr := zog.getdispatch_table
  par_pasm_addr := zog.getzog
  par_zog_mbox_addr := @zog_mbox_command
#ifdef USE_JCACHED_MEMORY
  par_vm_mbox_addr := vm_mbox
#endif
#ifdef USE_VIRTUAL_MEMORY
  par_vm_mbox_addr := @vm_mbox
#endif

  'vm_initSDCard
  ser.stop
  zog.start(@par_zpu_memory_addr)                           'Start ZOG with parameter block
  service_io                                                'Go any monitor ZPU I/O

#ifdef USE_SDCARD

#define DEBUG_SD
'#define DEBUG_SD_DATA

PUB load_bytecode | bytes, i, ramaddr
  ser.str (string("Opening ZPU image '"))
  ser.str (@filename)
  ser.str (string("'...",))
  err := \sd.popen(@filename, "r")
  ser.hex(err, 8)
  crlf

  ser.str (string("Reading image...",))

  bytes := 1
  ramaddr := 0
  repeat while bytes > 0
    bytes := sd.pread(@diskbuff, 512)
    if bytes > 0
      err := sd.stopSDCard                                'Stop SD card, but do not stop cog
      repeat i from 0 to bytes - 1
        vm_writeByte(ramaddr, diskbuff[i])                  'To virtual memory.
        ramaddr++
      err := \sd.initSDCard(spiDO,spiClk,spiDI,spiINC)     'Init the SD
  ser.dec (ramaddr)
  ser.str (string(" bytes"))
  crlf

  err := sd.stopSDCard                                  'stop SD card, but do not stop cog
  CheckError(err)

  ser.str (string("Clearing bss: ",))
  repeat i from ramaddr to zpu_memory_size-1
    vm_writeByte(i, 0)
    ifnot i & ($800-1)
      ser.tx(".")
  crlf
  par_initial_pc := 0

PUB write_bytecode_to_flash | bytes, i, ramaddr
  ser.str (string("Opening ZPU image '"))
  ser.str (@filename)
  ser.str (string("'...",))
  err := \sd.popen(@filename, "r")
  ser.hex(err, 8)
  crlf

  ser.str (string("Reading image...",))

  bytes := 1
  ramaddr := 0
  err := sd.stopSDCard                                'Stop SD card, but do not stop cog
  vm_EraseFlashBlock(ramaddr)
  err := \sd.initSDCard(spiDO,spiClk,spiDI,spiINC)     'Init the SD
  repeat while bytes > 0
    bytes := sd.pread(@diskbuff, 512)
    if bytes > 0
      err := sd.stopSDCard                                'Stop SD card, but do not stop cog
      vm_WriteFlash(ramaddr, @diskbuff, bytes)
      ramaddr += bytes
      if (ramaddr & $00000fff) == 0
        vm_EraseFlashBlock(ramaddr)
      err := \sd.initSDCard(spiDO,spiClk,spiDI,spiINC)     'Init the SD
  ser.dec (ramaddr)
  ser.str (string(" bytes"))
  crlf

  err := sd.stopSDCard                                  'stop SD card, but do not stop cog
  CheckError(err)

PRI swap(value) | swapped
  return value
  swapped :=  value              << 24
  swapped |= (value & $0000ff00) << 8
  swapped |= (value & $00ff0000) >> 8
  swapped |=  value              >> 24
  return swapped

PUB start_bytecode_in_flash | data_image, data_start, data_end, bss_start, bss_end
  sd.stopSDCard

  ' load the image header
  par_initial_pc     := swap(vm_readLong($100000))
  data_image := swap(vm_readLong($100004))
  data_start := swap(vm_readLong($100008))
  data_end   := swap(vm_readLong($10000c))
  bss_start  := swap(vm_readLong($100010))
  bss_end    := swap(vm_readLong($100014))
'{
  ser.str(string("start:      "))
  ser.hex(par_initial_pc, 8)
  ser.crlf
  ser.str(string("data_image: "))
  ser.hex(data_image, 8)
  ser.crlf
  ser.str(string("data_start: "))
  ser.hex(data_start, 8)
  ser.crlf
  ser.str(string("data_end:   "))
  ser.hex(data_end, 8)
  ser.crlf
  ser.str(string("bss_start:  "))
  ser.hex(bss_start, 8)
  ser.crlf
  ser.str(string("bss_end:    "))
  ser.hex(bss_end, 8)
  ser.crlf
'}
  ' initialize .data
  repeat while data_start < data_end
    vm_writeLong(data_start, vm_readLong(data_image))
    data_image += 4
    data_start += 4

  ' clear .bss
  repeat while bss_start < bss_end
    vm_writeLong(bss_start, 0)
    bss_start += 4

PUB prime_vm | vbyte, count, ramaddr
  ser.str (string("Priming VM...", 13))
  repeat count from 1 to 2
    repeat ramaddr from $FC00 to $FFFF
      vbyte := vm_readByte(ramaddr)
  repeat count from 1 to 1000
    repeat ramaddr from $05b9 to $05d5
      vbyte := vm_readByte(ramaddr)
  ser.str (string("Done", 13))

PRI dump_virtual | ad, b, i, e
  ser.str(string("Reading virtual memory...",13,10))
  i := 0
  e := 0
  repeat ad from 0 to 127 'zpu_memory_size - 1 step 1
    if (i & 15) == 0
      ser.hex(ad,4)
      ser.tx(":")
      ser.tx(" ")
    b:=vm_readByte(ad ^ %11)                             'XOR here is an endiannes fix
    ser.hex(b,2)
    ser.tx($20)
    i := i + 1
    if i == 16
      ser.tx(13)
      ser.tx(10)
      i := 0

pub check_bytecode | bytes, i, ramaddr, mbyte, error, sbytes, secs

  err := sd.stop

  secs := 2
  crlf
  ser.str (string("Waiting "))
  ser.dec(secs)
  ser.str (string(" seconds before program check...",))
  waitcnt(clkfreq*secs+cnt)
  crlf
  crlf

  ser.str (string("Restarting SD driver...",))
  err := sd.startSD(@ioControl)                         'Start the SD routines
  ser.hex(err, 8)
  crlf

  ser.str (string("Remounting SD...",))
  err := \sd.mount_explicit(spiDO,spiClk,spiDI,spiINC)    'Mount the SD
  ser.hex(err, 8)
  crlf

  ser.str (string("Checking image...",))

  err := \sd.popen(@filename, "r")
  error := 0
  ramaddr := 0
  bytes := 1
  sbytes~
  repeat while bytes > 0
    bytes := sd.pread(@diskbuff, 512)
    if (bytes > 0)
      sbytes += bytes
      err := sd.stopSDCard                                'stop SD card, but do not stop cog
      repeat i from 0 to bytes - 1
        mbyte := vm_readByte(ramaddr)
        if mbyte <> diskbuff[i]                           'Check virtual memory.
          ser.str(string($d,$a,"Program load error: "))
          ser.hex(ramaddr,8)
          ser.str(string(" expected : "))
          ser.hex(diskbuff[i],2)
          ser.str(string(" received : "))
          ser.hex(mbyte,2)
          ser.str(string(" ("))
          ser.dec(ramaddr)
          ser.str(string(")"))
          error++
        ramaddr++
      err := \sd.initSDCard(spiDO,spiClk,spiDI,spiINC)     'Init the SD

  ser.str(string(" "))
  ser.dec(sbytes)
  ser.str(string(" Bytes Checked."))
  if error
    ser.str (string($d,$a,"Program Load Failed."))
  else
    ser.str (string($d,$a,"Program Load OK."))
  crlf

  err := sd.stopSDCard                                  'stop SD card, but do not stop cog
  CheckError(err)

#endif


PRI service_io
'ZPU I/O simulation service loop
  repeat
    case zog_mbox_command
      io_cmd_break:                                     'CPU break
        on_break
      io_cmd_in:                                        'An input operation
        on_input
      io_cmd_out:                                       'An output operation
        on_output
      io_cmd_syscall:
        on_syscall
#ifdef CYCLE_COUNT
    ser.str(string("Cycles = "))
    ser.dec(debug)
    crlf
#endif

PRI on_input
  'ser.str(string("in: "))
  'ser.hex(zog_mbox_port, 8)
  'crlf
  case zog_mbox_port
    UART_TX_PORT:
      zog_mbox_data := $100
    UART_RX_PORT:
      zog_mbox_data := ser.rx + $100
    other:
      ser.str(string("Read from unknown input port "))
      ser.hex(zog_mbox_port, 8)
      crlf
  zog_mbox_command := 0

PRI on_output
  'ser.str(string("out: "))
  'ser.hex(zog_mbox_port, 8)
  'crlf
  case zog_mbox_port
    UART_TX_PORT:
      ser.tx(zog_mbox_data)
    other:
      ser.str(string("Write to unknown output port "))
      ser.hex(zog_mbox_port, 8)
      ser.str(string(" : "))
      ser.hex(zog_mbox_data, 8)
      crlf
  zog_mbox_command := 0

PRI on_break
'Handle CPU break condition
  if step_count == 0
    crlf
    ser.str(string("pc       op sp       tos      reason"))
    crlf
  if print_regs <> 0
    repeat

'  repeat while ser.rx <> $20                           'Wait for 'space'
  zog_mbox_command := 0
  if step_count++ == 1000
    ser.str(string("Done."))
    crlf
    repeat

PRI printStackFrame | buff, i
'Labels as per C read()/write() parameters.
  ser.str(string("Return address="))
  ser.hex(vm_readLong(framep + 0), 8)
  crlf
  ser.str(string("errno address="))
  ser.hex(vm_readLong(framep + 4), 8)
  crlf
  ser.str(string("syscall ID="))
  ser.hex(vm_readLong(framep + 8), 8)
  crlf
  ser.str(string("fd="))
  ser.hex(vm_readLong(framep + 12), 8)
  crlf
  ser.str(string("buff="))
  ser.hex(vm_readLong(framep + 16), 8)
  crlf
  ser.str(string("length="))
  ser.hex(vm_readLong(framep + 20), 8)
  crlf

PRI on_syscall | p
'Handle CPU SYSCALL instruction
'  Example C call:
'  result=_syscall(&t, SYS_write, fd, buf, nbytes);
  'Set a frame pointer into the ZPU stack (but HUB realtive)
  framep := zog_mbox_sp
  'printStackFrame
  'Perform system call function given by the ID param
  case vm_readLong(framep + 8)
    SYS_exit:
      _exit
    SYS_open:
      _open
    SYS_close:
      _close
    SYS_read:
      _read
    SYS_write:
      _write
    SYS_lseek:
      _lseek
    SYS_unlink:
      _unlink
    SYS_getpid:
      _getpid
    SYS_kill:
      _kill
    SYS_fstat:
      _fstat
    SYS_argvlen:
      _argvlen
    SYS_argv:
      _argv
    SYS_chdir:
      _chdir
    SYS_stat:
      _stat
    SYS_chmod:
      _chmod
    SYS_utime:
      _utime
    SYS_time:
      _time
  zog_mbox_command := 0

PRI _exit
  ser.str(string("_exit ?"))
  repeat

PRI _open
'Open a file
'In C, result=_syscall(&t, SYS_open, name, flags, mode);
  ser.str(string("_open ?"))
  repeat

PRI _close
'Close a file
'In C, result=_syscall(&t, SYS_close, fileno);
  ser.str(string("_close ?"))
  repeat

PRI _read | i, f, p, c, c2, retval
'Read nbytes of fd to buffer
'In C, result=_syscall(&t, SYS_read, fd, buf, nbytes);
  'FIXME why not just add an offset to sp?
  f := vm_readLong(framep + 16)                   'Get buff ptr from ZPU stack
  p := f
  c := vm_readLong(framep + 20)
  case vm_readLong(framep + 12)                         'Check file descriptor
    STDIN_FILENO, STDERR_FILENO:                        'Console input?
      repeat i from 0 to c - 1   'Input nbytes
        vm_writeByte(p ^ %11,ser.rx)                    'XOR here is an endiannes fix
        p++
      retval := p - f
#ifdef USE_SDCARD
    SECTORIO_FILENO:
#ifdef DEBUG_SD
      ser.str(string("_read: "))
      ser.hex(file_offset,8)
      ser.str(string(" "))
      ser.hex(c,8)
#endif
      repeat while c > 0
        c2 := c <# 512
        'sd.initSDCard(spiDO,spiClk,spiDI,spiINC)
        'if sd.readSDCard(file_offset>>9, @diskbuff, c2) <> 0
        if vm_readSDCard(file_offset>>9, @diskbuff, c2) <> 0
#ifdef DEBUG_SD
          ser.str(string(" -- failed"))
          ser.crlf
#endif
          return -1
        'sd.stopSDCard                                'Stop SD card, but do not stop cog
#ifdef DEBUG_SD_DATA
        ser.crlf
        ser.str(string("offset "))
        ser.hex(file_offset,8)
#endif
        repeat i from 0 to c2 - 1
          vm_writeByte(p ^ %11,diskbuff[i])
#ifdef DEBUG_SD_DATA
          if (i // 16) == 0
            ser.crlf
            ser.hex(file_offset+i,8)
            ser.str(string(":"))
          ser.tx(" ")
          ser.hex(diskbuff[i],2)
#endif
          p++
        file_offset += c2
        c -= c2
      retval := p - f
#ifdef DEBUG_SD_DATA
      ser.crlf
#endif
#ifdef DEBUG_SD
      ser.str(string(" -> "))
      ser.hex(retval,8)
      ser.crlf
#endif
#endif
    other:
      'FIXME: Other file and device input here
      retval := -1
  'FIXME errno
  vm_writeLong(0,retval)                         'Return value via _mreg (aka R0) at ZPU address 0 !

PRI _write | i, p, f, c, c2, retval
'Write nbytes of buffer to fd
'In C, result=_syscall(&t, SYS_write, fd, buf, nbytes);
  'FIXME why not just add an offset to sp?
  f := vm_readLong(framep + 16)                   'Get buff ptr from ZPU stack
  p := f
  c := vm_readLong(framep + 20)
  case vm_readLong(framep + 12)                         'Check file descriptor
    STDOUT_FILENO,STDERR_FILENO:                        'Console output?
      repeat i from 0 to c - 1   'Output nbytes
        ser.tx(vm_readByte(p ^ %11))                    'XOR here is an endiannes fix
        p++
      retval := p - f
#ifdef USE_SDCARD
    SECTORIO_FILENO:
#ifdef DEBUG_SD
      ser.str(string("_write: "))
      ser.hex(file_offset,8)
      ser.str(string(" "))
      ser.hex(c,8)
#endif
      repeat while c > 0
        c2 := c <# 512
#ifdef DEBUG_SD_DATA
        ser.crlf
        ser.str(string("offset "))
        ser.hex(file_offset,8)
#endif
        repeat i from 0 to c2 - 1
          diskbuff[i] := vm_readByte(p ^ %11)
#ifdef DEBUG_SD_DATA
          if (i // 16) == 0
            ser.crlf
            ser.hex(file_offset+i,8)
            ser.str(string(":"))
          ser.tx(" ")
          ser.hex(diskbuff[i],2)
#endif
          p++
        'sd.initSDCard(spiDO,spiClk,spiDI,spiINC)
        'if sd.writeSDCard(file_offset>>9, @diskbuff, c2) <> 0
        if vm_writeSDCard(file_offset>>9, @diskbuff, c2) <> 0
#ifdef DEBUG_SD
          ser.str(string(" -- failed"))
          ser.crlf
#endif
          return -1
        'sd.stopSDCard                                'Stop SD card, but do not stop cog
        file_offset += c2
        c -= c2
      retval := p - f
#ifdef DEBUG_SD_DATA
      ser.crlf
#endif
#ifdef DEBUG_SD
      ser.str(string(" -> "))
      ser.hex(retval,8)
      ser.crlf
#endif
#endif
    other:
      'FIXME: Other file and device output here
      retval := -1
  'FIXME errno
  vm_writeLong(0,retval)                         'Return value via _mreg (aka R0) at ZPU address 0 !

PRI _lseek | retval
  'FIXME why not just add an offset to sp?
  case vm_readLong(framep + 12)                         'Check file descriptor
    STDOUT_FILENO,STDERR_FILENO:                        'Console output?
      retval := -1
#ifdef USE_SDCARD
    SECTORIO_FILENO:
      if (vm_readLong(framep + 20) == 0)
        file_offset := vm_readLong(framep + 16)
        ser.str(string("_lseek: "))
        ser.hex(file_offset,8)
        ser.crlf
        retval := file_offset
      else
        retval := -1
#endif
    other:
      'FIXME: Other file and device output here
      retval := -1
  'FIXME errno
  vm_writeLong(0,retval)                         'Return value via _mreg (aka R0) at ZPU address 0 !

PRI _unlink
  ser.str(string("_unlink ?"))
  repeat

PRI _getpid
  ser.str(string("_getpid ?"))
  repeat

PRI _kill
  ser.str(string("_kill ?"))
  repeat

PRI _fstat
  vm_writeLong(0,0)                             'Return value via _mreg (aka R0) at ZPU address 0 !

PRI _argvlen
  ser.str(string("_argvlen ?"))
  repeat

PRI _argv
  ser.str(string("_argv ?"))
  repeat

PRI _chdir
  ser.str(string("_chdir ?"))
  repeat

PRI _stat
  ser.str(string("_stat ?"))
  repeat

PRI _chmod
  ser.str(string("_chmod ?"))
  repeat

PRI _utime
  ser.str(string("_utime ?"))
  repeat

PRI _time
  ser.str(string("_time ?"))
  repeat

PRI print_regs | i, p, op
'Print the ZPU registers
  ser.hex(zog_mbox_pc, 8)
  ser.str(string(" "))
  op := vm_readByte(zog_mbox_pc ^ %11)                  'XOR here is an endiannes fix
  ser.hex(op, 2)
  ser.str(string(" "))
  ser.hex(zog_mbox_sp, 8)
  ser.str(string(" "))
  ser.hex(zog_mbox_tos, 8)
  ser.str(string(" "))

  case op
    $00:
      ser.str(string("BREAKPOINT"))
      result := 1
    $01,$03,$0E,$0F:
      ser.str(string("ILLEGAL OP"))
      result := 1
    $20,$21:
      ser.str(string("NOT_IMPLEMENTED"))
      result := 1
    other:
      result := 0
  crlf

'  p := 0 - 8*4
'  repeat i from 0 to 7
'    ser.hex(p + 4*i, 8)
'    ser.tx($20)
'    ser.hex(LONG[@zpu_memory[(p {& zpu_memory_mask})+ 4*i]], 8)
'    crlf
'  crlf

pub vm_readLong(madr)
#ifdef USE_VIRTUAL_MEMORY
    return vm.rdvlong(madr)
#elseifdef USE_JCACHED_MEMORY
    long[vm_mbox][0] := (madr&!1) | cm#READ_CMD
    repeat while long[vm_mbox][0]
    madr &= cm#LINELEN-1
    return long[long[vm_mbox][1]+madr]
#elseifdef USE_HUB_MEMORY
    return long[zpu_memory+madr]
#endif

pub vm_writeLong(madr,val)
#ifdef USE_VIRTUAL_MEMORY
    vm.wrvlong(madr, val)
#elseifdef USE_JCACHED_MEMORY
    long[vm_mbox][0] := (madr&!1) | cm#WRITE_CMD
    repeat while long[vm_mbox][0]
    madr &= cm#LINELEN-1
    long[long[vm_mbox][1]+madr] := val
#elseifdef USE_HUB_MEMORY
    long[zpu_memory+madr] := val
#endif

pub vm_readByte(madr)
#ifdef USE_VIRTUAL_MEMORY
    return vm.rdvbyte(madr)
#elseifdef USE_JCACHED_MEMORY
    long[vm_mbox][0] := (madr&!1) | cm#READ_CMD
    repeat while long[vm_mbox][0]
    madr &= cm#LINELEN-1
    return byte[long[vm_mbox][1]+madr]
#elseifdef USE_HUB_MEMORY
    return byte[zpu_memory+madr]
#endif

pub vm_writeByte(madr,val)
#ifdef USE_VIRTUAL_MEMORY
    vm.wrvbyte(madr, val)
#elseifdef USE_JCACHED_MEMORY
    long[vm_mbox][0] := (madr&!1) | cm#WRITE_CMD
    repeat while long[vm_mbox][0]
    madr &= cm#LINELEN-1
    byte[long[vm_mbox][1]+madr] := val
#elseifdef USE_HUB_MEMORY
    byte[zpu_memory+madr] := val
#endif

pub vm_eraseFlashBlock(madr)
  ser.str(string("Erasing "))
  ser.hex(madr, 8)
  long[vm_mbox][0] := cm#ERASE_BLOCK_CMD | (madr << 8)
  repeat while long[vm_mbox][0] <> 0
  ser.str(string("-> "))
  ser.hex(long[vm_mbox][1], 8)
  ser.crlf
  return long[vm_mbox][1]

pub vm_writeFlash(madr,buf,count) | pbuf, pcnt, paddr
  ser.str(string("Writing "))
  ser.hex(madr, 8)
  pbuf := buf
  pcnt := count
  paddr := madr
  long[vm_mbox][0] := cm#WRITE_DATA_CMD | (@pbuf << 8)
  repeat while long[vm_mbox][0] <> 0
  ser.str(string("-> "))
  ser.hex(long[vm_mbox][1], 8)
  ser.crlf
  return long[vm_mbox][1]

PUB vm_initSDCard
  ser.str(string("Init SD card "))
  long[vm_mbox][0] := cm#SD_INIT_CMD
  repeat while long[vm_mbox][0] <> 0
  ser.str(string("-> "))
  ser.hex(long[vm_mbox][1], 8)
  ser.crlf
  return long[vm_mbox][1]

PUB vm_readSDCard(addr,buffer,count) | pbuf, pcnt, paddr
  ser.str(string("Read SD card "))
  ser.hex(addr, 8)
  pbuf := buffer
  pcnt := count
  paddr := addr
  long[vm_mbox][0] := cm#SD_READ_CMD | (@pbuf << 8)
  repeat while long[vm_mbox][0] <> 0
  ser.str(string("-> "))
  ser.hex(long[vm_mbox][1], 8)
  ser.crlf
  return long[vm_mbox][1]

PUB vm_writeSDCard(addr,buffer,count) | pbuf, pcnt, paddr
  ser.str(string("Write SD card "))
  ser.hex(addr, 8)
  pbuf := buffer
  pcnt := count
  paddr := addr
  long[vm_mbox][0] := cm#SD_WRITE_CMD | (@pbuf << 8)
  repeat while long[vm_mbox][0] <> 0
  ser.str(string("-> "))
  ser.hex(long[vm_mbox][1], 8)
  ser.crlf
  return long[vm_mbox][1]

PRI crlf
  ser.tx(13)
  ser.tx(10)

PRI CheckError(r)
  if r < 0
    ser.str(string("Error: "))
    ser.hex(r,2)
    ser.str(string(" HALTED."))
    crlf
'    sd.stop
    repeat                                             '<== LOOP HERE

'---------------------------------------------------------------------------------------------------------

'---------------------------------------------------------------------------------------------------------
'The End.

